import setuptools

setuptools.setup(
    name="fasteasySD",
    version="0.0.3",
    license_files = ('LICENSE.txt',),
    author="WaGaNaWa",
    author_email="m4a1carbin4@gmail.com",
    description="simple_fast_easy_sd",
    long_description=open('README.md').read(),
    url="https://github.com/m4a1carbin4/fasteasysd",
    packages=setuptools.find_packages(),
)